#!/usr/bin/env python

"""
The simplify package for processing Python source code.

Copyright (C) 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

To use this module, an importer should be constructed and the load_from_file
method used. Here, the standard path for module searching is employed:

importer = Importer(sys.path)
importer.load_from_file(builtins)
importer.load_from_file(filename, builtins)
"""

from simplify.simplified import *
import simplify.ast
import simplify.fixnames
import simplify.annotate
import simplify.fixinstances

class Importer:

    "An import machine, searching for and loading modules."

    def __init__(self, path=None, annotate=1, fixinstances=1):

        """
        Initialise the importer with the given search 'path' - a list of
        directories to search for Python modules. If the optional 'annotate'
        parameter is set to a false value (unlike the default), no annotation
        will be performed.
        false value (unlike the default), no instance fixing will be performed.
        """

        self.path = path or [os.getcwd()]
        self.path.append(libdir)
        self.annotate = annotate
        self.modules = {}

    def get_modules(self):

        "Return all modules known to the importer."

        return self.modules.values()

    def find_in_path(self, name):

        """
        Find the given module 'name' in the search path, returning None where no
        such module could be found, or a 2-tuple from the 'find' method
        otherwise.
        """

        for d in self.path:
            m = self.find(d, name)
            if m: return m
        return None

    def find(self, d, name):

        """
        In the directory 'd', find the given module 'name', where 'name' can
        either refer to a single file module or to a package. Return None if the
        'name' cannot be associated with either a file or a package directory,
        or a 2-tuple from '_find_package' or '_find_module' otherwise.
        """

        m = self._find_package(d, name)
        if m: return m
        m = self._find_module(d, name)
        if m: return m
        return None

    def _find_module(self, d, name):

        """
        In the directory 'd', find the given module 'name', returning None where
        no suitable file exists in the directory, or a 2-tuple consisting of
        None (indicating that no package directory is involved) and a filename
        indicating the location of the module.
        """

        name_py = name + os.extsep + "py"
        filename = self._find_file(d, name_py)
        if filename:
            return None, filename
        return None

    def _find_package(self, d, name):

        """
        In the directory 'd', find the given package 'name', returning None
        where no suitable package directory exists, or a 2-tuple consisting of
        a directory (indicating the location of the package directory itself)
        and a filename indicating the location of the __init__.py module which
        declares the package's top-level contents.
        """

        filename = self._find_file(d, name)
        if filename:
            init_py = "__init__" + os.path.extsep + "py"
            init_py_filename = self._find_file(filename, init_py)
            if init_py_filename:
                return filename, init_py_filename
        return None

    def _find_file(self, d, filename):

        """
        Return the filename obtained when searching the directory 'd' for the
        given 'filename', or None if no actual file exists for the filename.
        """

        filename = os.path.join(d, filename)
        if os.path.exists(filename):
            return filename
        else:
            return None

    def load(self, name, builtins=None, alias=None):

        """
        Load the module or package with the given 'name' and using the specified
        'builtins'. Return an Attribute object referencing the loaded module or
        package, or None if no such module or package exists.
        """

        if self.modules.has_key(name):
            return Attribute(None, self.modules[name])

        # Split the name into path components, and try to find the uppermost in
        # the search path.

        path = name.split(".")
        m = self.find_in_path(path[0])
        if not m:
            return None # NOTE: Import error.
        d, filename = m

        # Either acquire a reference to an already-imported module, or load the
        # module from a file.

        if self.modules.has_key(path[0]):
            top = module = self.modules[path[0]]
        else:
            top = module = self.load_from_file(filename, builtins, path[0])

        # For hierarchical names, traverse each path component...

        if len(path) > 1:
            path_so_far = path[:1]
            for p in path[1:]:
                path_so_far.append(p)

                # Find the package or module concerned.

                m = self.find(d, p)
                if not m:
                    return None # NOTE: Import error.
                d, filename = m
                module_name = ".".join(path_so_far)

                # Either reference an imported module or load one from a file.

                if self.modules.has_key(module_name):
                    submodule = self.modules[module_name]
                else:
                    submodule = self.load_from_file(filename, builtins, module_name)

                # Store the submodule within its parent module.

                module.namespace[p] = [Attribute(None, submodule)]
                module = submodule

        # Return either the deepest or the uppermost module.

        if alias:
            return Attribute(None, module)
        else:
            return Attribute(None, top)

    def load_from_file(self, name, builtins=None, module_name=None):

        """
        Load the module with the given 'name' (which may be a full module path),
        using the optional 'builtins' to resolve built-in names.
        """

        if module_name is None:
            if builtins is None:
                module_name = "__builtins__"
            else:
                module_name = "__main__"

        # Simplify, fix names, and annotate the module.

        module = simplify.ast.simplify(name, builtins is None, module_name)
        simplify.fixnames.fix(module, builtins)
        if self.annotate:
            simplify.annotate.annotate(module, builtins, self)

        # Record the module.

        self.modules[module_name] = module
        return module

    def fix_instances(self):

        "Fix instances for all modules loaded by this importer."

        for module in self.get_modules():
            simplify.fixinstances.fix_structures(module)
        for module in self.get_modules():
            simplify.fixinstances.fix(module)

# vim: tabstop=4 expandtab shiftwidth=4
